/*
 * Copyright (c) Sun Microsystems, Inc. 1993-1998 All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SMCC Technology
 *      Development Group at Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" 
 * without express or implied warranty of any kind. 
 *  
 * These notices must be retained in any copies of any part of this software.
 */

#pragma ident "@(#)cbq.c  1.39     98/05/13 SMI"

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/strsun.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/dlpi.h>
#include <sys/stropts.h>
#include <sys/strlog.h>
#include <sys/tihdr.h>
#include <sys/time.h>
#include <sys/tiuser.h>
#include <sys/kstat.h>
#include <sys/kmem.h>

#include <sys/socket.h>
#include <net/route.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <inet/common.h>
#include <inet/mi.h>
#include <inet/ip.h>
#include <inet/ip_multi.h>
#include <inet/ip_ire.h>
#include <netinet/ip_mroute.h>
#include <sys/ethernet.h>

#include <netinet/rm_class.h>
#include <netinet/pktsched.h>
#include <netinet/cbq.h>

/*
 * Local Data structures.
 */


static kmutex_t		cbq_lock_g;
static cbq_state_t	*cbq_ill_head_g = NULL;
static cbq_state_t	*cbq_device_g = NULL;
static cbq_state_t	*cbq_ctl_g = NULL;
static void		*cbq_head_g = NULL;

/*
 * Module Global data.
 */
int			cbq_debug = 0;

/*
 * Forward Declarations.
 */
static int		cbq_open(queue_t *, dev_t *, int, int, cred_t *);
static int		cbq_module_open(queue_t *, dev_t *, int, int, cred_t *);
static int		cbq_close(queue_t *);
static int		cbq_wput(queue_t *, mblk_t *);
static int		cbq_wsrv(queue_t *);
static int		cbq_rput(queue_t *, mblk_t *);
static int		cbq_rsrv(queue_t *);
static void		cbq_ioctl(queue_t *, mblk_t *);
static void		cbq_ioc(queue_t *, mblk_t *);
static int		cbq_proto(queue_t *, mblk_t *);
static void		cbq_pktsched(void *arg, int);
static void		cbq_esballoc_free(void *arg);
static void		cbqfree(void *arg, int);
static int		cbq_deque(queue_t *);

#ifdef CBQPERF
static int		cbq_getnperf(struct cbq_getnperf *getnpfp);
static int		cbq_getperf(struct cbq_getperf *getpfp);
#endif

static int		cbq_set_enable(struct cbq_interface *, int);
static void		cbq_satisfied(cbq_module_t *cbqp);
static int		cbq_getstats(struct cbq_getstats *gs);
static int		_cbq_getstats(struct cbq_getstats *gs,
				cbq_module_t *cbqp);
static int		cbq_add_filter(struct cbq_add_filter *);
static int		cbq_delete_filter(struct cbq_delete_filter *);
static int		cbq_add_class(struct cbq_add_class *);
static int		cbq_delete_class(struct cbq_delete_class *);
static int		cbq_clear_hierarchy(struct cbq_interface *);

static int		cbq_clear_interface(cbq_module_t *);
static void		cbq_dealloc_filter(cbq_module_t *, k_filter_cb_t *);
static int		is_a_parent(cbq_module_t *, struct rm_class *);
static k_filter_cb_t 	*cbq_filth_to_filtp(cbq_module_t *, u_long, int *);
static u_long		cbq_alloc_class_handle(cbq_module_t *, int *);
static void		cbq_free_class_handle(cbq_module_t *, u_long);
static struct rm_class  *cbq_clh_to_clp(cbq_module_t *, u_long, int *);

static int		cbq_ill_init(queue_t *, cbq_state_t *);
cbq_state_t			*cbq_ill_lookup_on_name(char *, int);
static int		cbq_ill_delete_entry(cbq_state_t *dcqp);
static int		is_a_parent_class(cbq_module_t *cbqp,
				struct rm_class *cl);

/*
 * Local Data structures.
 */
static struct module_info info = {
	0x4351, "cbq", 1, INFPSZ, CBQ_HI_WATER, CBQ_LO_WATER
};

static struct qinit rinit = {
	cbq_rput, cbq_rsrv, cbq_open, cbq_close, NULL, &info, NULL
};

static struct qinit winit = {
	cbq_wput, cbq_wsrv, cbq_open, cbq_close, NULL, &info, NULL
};

/*
 * Global data structures.
 */
struct streamtab cbqinfo = {
	&rinit, &winit, (struct qinit *)NULL, (struct qinit *)NULL
};

/*
 * Streams Module/Driver local functions.
 */

static int
cbq_pktsched_init(cbq_module_t *cbqp)
{
	pkt_sched_cb_t	cb;

	cb.pktsched_cb = cbq_pktsched;
	cb.pktsched_cb_arg = (void *)cbqp;

	if (!cbqp->is_pktsched_inited) {
		cbqp->is_pktsched_destroyed = 0;
		return (pktsched_init(cbqp->cbq_q, &cb,
			&cbqp->is_pktsched_inited,
			&cbqp->cbq_lock, &cbqp->cv));
	}
}

static int
cbq_pktsched_destroy(cbq_module_t *cbqp)
{
	if (cbqp->is_pktsched_inited) {
		cbqp->is_pktsched_inited = 0;

		return (pktsched_teardown(cbqp->cbq_q,
			&cbqp->is_pktsched_destroyed,
			&cbqp->cbq_lock, &cbqp->cv));
	} else
		return (1);
}

int
cbq_ddi_init(dev_info_t *devi)
{
#ifdef notdef
	cbq_device_g = kmem_zalloc(sizeof(cbq_state_t), KM_SLEEP);
	if (!cbq_device_g) {
		cmn_err(CE_CONT, "cbq_ddi_init: kmem_zalloc FAILED\n"); 
		return (0);
	}

	cbq_device_g->cbq_id = CBQ_DEVICE;
	cbq_device_g->cbq.cbq_ds.dinfop = devi;

	cbq_ctl_g = kmem_zalloc(sizeof(cbq_state_t), KM_SLEEP);
	if (!cbq_ctl_g) {
		cmn_err(CE_CONT, "cbq_ddi_init: kmem_zalloc FAILED\n"); 
		if (cbq_device_g)
			kmem_free(cbq_device_g, sizeof(cbq_state_t));
		return (0);
	}

	cbq_ctl_g->cbq_id = CBQ_DEVICE;
	cbq_ctl_g->cbq.cbq_ds.dinfop = devi;
#endif

	mutex_init(&cbq_lock_g, (char *)NULL, MUTEX_DEFAULT, (void *)NULL);
	return (1);
}

int
cbq_ddi_destroy(void)
{
	mutex_destroy(&cbq_lock_g);
#ifdef notdef
	kmem_free(cbq_ctl_g, sizeof(cbq_state_t));
	kmem_free(cbq_device_g, sizeof(cbq_state_t));
#endif
	return (1);
}

/*
 * static int
 * cbq_open(queue_t *, dev_t *, int, int, cred_t *) - This routine
 *	handles both Module and driver open.  If we are doing a Module
 *	open, we are being pushed on the Network Interface stack.  If we
 *	doing a driver open, then /dev/cbq is being opened to perform
 *	ioctl's on various network interfaces.
 *
 * NOTE:  Device must remain open for all state to be in place. All
 *	link-sharing state is toredown on close.
 *
 *	Returns:	ENODEV, if we opened too many /dev/cbqs.
 *			EBUSY, if we the device is already open.
 *			0, if we are successful.
 */

static int
cbq_open(queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp)
{
	int		i, instance, err = 0;
	cbq_state_t	*cqp;
	char		*cp;

	cbq0dbg((CE_CONT,"cbq_open\n"));

	if (sflag == CLONEOPEN) {
		instance = getminor(*devp);
		cmn_err(CE_CONT, "cbq_open: instance is %d\n", instance);
	}

	qprocson(q);

	/*
	 * We are either opening as a device or a module.  In the device
	 * case, this is a CBQ client and we allocate a device structure.
	 * If it is a module open, then this is control stream access to the
	 * DLPI device.  We allocate a cbq_ill_t as the instance data in this
	 * case.
	 */

	/*
	 * Are we being opened as a module?  If so, we are being pushed
	 * onto the interface ahead of IP.
	 */

	if (sflag == MODOPEN) {
		err = cbq_module_open(q, devp, flag, sflag, credp);
	} else {
		mutex_enter(&cbq_lock_g);

		cqp = (cbq_state_t *)q->q_ptr;
		if (cqp == NULL) {
			/*
		 	 * Open the cbq device
		 	 */
			if (err = mi_open_comm(&cbq_head_g,
					sizeof (cbq_state_t), q, devp,
					flag, sflag, credp)) {
				mutex_exit(&cbq_lock_g);
				goto cbq_open_exit;
			}

			cqp = (cbq_state_t *)q->q_ptr;
		}

		/*
	 	 * Check to see if the device is already open.
	 	 */
		if (cqp->cbq.cbq_ds.cbq_open == true) {
			err = EBUSY;
			mutex_exit(&cbq_lock_g);
			goto cbq_open_exit;
		}

		cqp->cbq_id = CBQ_DEVICE;
		cqp->cbq.cbq_ds.cbq_open = true;
		cqp->cbq.cbq_ds.cbq_q = q;

		if (!CBQ_ISCTL(getminor(*devp))) {
			cqp->cbq.cbq_ds.cbq_minor = CBQ_MINOR_DEVICE;
			cbq_ctl_g = cqp;
		} else {
			cqp->cbq.cbq_ds.cbq_minor = CBQ_MINOR_CTL;
			cbq_device_g = cqp;
		}

		mutex_exit(&cbq_lock_g);
	}

cbq_open_exit:
	if (err)
		cmn_err(CE_CONT, "cbq_open: error is %d\n", err);
	return (err);
}

static int
cbq_module_open(queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp)
{
	queue_t		*downstreamq;
	cbq_state_t	*cqp;
	cbq_module_t	*cbqp;
	int		i, err = 0;

	mutex_enter(&cbq_lock_g);

	downstreamq = WR(q)->q_next;
	if (!downstreamq) {
		err = ENXIO;
		goto cbq_module_open_exit;
	}

	if (err = mi_open_comm(&cbq_head_g, sizeof (cbq_state_t), q, devp,
			flag, sflag, credp))
		goto cbq_module_open_exit;

	cqp = (cbq_state_t *)WR(q)->q_ptr;
	cqp->cbq_id = CBQ_MODULE;
	cbqp = &cqp->cbq.cbq_ms.module;

	/*
	 * Initialize the Module state structure.
	 */
	if (err = cbq_ill_init(WR(q), cqp)) {
		mi_close_comm(&cbq_head_g, q);
		goto cbq_module_open_exit;
	}

	/*
	 * Initialize the module structure.
	 */
	mutex_init(&cqp->cbq.cbq_ms.module.cbq_lock, (char *)NULL,
		MUTEX_DEFAULT, NULL);
	cv_init(&cqp->cbq.cbq_ms.module.cv, (char *)NULL,
		CV_DEFAULT, NULL);

	cbqlock(WR(q));
	cqp->cbq.cbq_ms.module.cbq_q = q;

	cbqp->cbqfi = 0;
	cbqp->cbqfo = 0;
	for (i = 0; i < RM_MAXQUEUED; i++)
		cqp->cbq.cbq_ms.module.cbq_freeit[i] = NULL;
	cqp->cbq.cbq_ms.module.cbq_retn.free_func = cbq_esballoc_free;
	cqp->cbq.cbq_ms.module.cbq_retn.free_arg =
		(char *)&cqp->cbq.cbq_ms.module;

	cqp->cbq.cbq_ms.module.cbq_timeout_id = 0;
	cqp->cbq.cbq_ms.module.q_full = 0;
	cqp->cbq.cbq_ms.module.qlen = 0;

	cbqp->cbq_class_tbl = (struct rm_class **)kmem_zalloc(
		sizeof (struct rm_class *) * CBQ_MAX_CLASSES, KM_SLEEP);
	if (!cbqp->cbq_class_tbl) {
		err = ENOMEM;
		mi_close_comm(&cbq_head_g, q);
		goto cbq_module_open_exit;
	}

#ifdef CBQPERF
	cbqp->perf = NULL;
	cbqp->nperf = 0;
#endif
	cbqp->is_pktsched_inited = 0;
	cbqp->is_pktsched_destroyed = 0;
	cbqp->use_esballoc = 0;
	cbqp->mac_addr_sz = 0;
	cbqp->cbq_filter_debug = 0;

	for (i = 0; i < CBQ_FILTER_TBL_SZ; i++)
		cbqp->filters[i] = NULL;

	cbqp->threshold = CBQ_VIRTUAL_IF_THRESH;

	cbqp->cbq_stats_enabled = 0;
	cbqp->cbq_stats_timeout_id = 0;
	cbqunlock(WR(q));

cbq_module_open_exit:
	mutex_exit(&cbq_lock_g);
	return (err);
}

/*
 * static int
 * cbq_close(queue_t *) - This functions closes the cbq Module or driver.
 *
 *	Returns:	0, on success.
 */

static int
cbq_close(queue_t *q)
{
	cbq_state_t	*mcqp, *cqp;
	cbq_module_t	*cbqp;
	cbq_device_t	*devp;
	int		result;

	cbq0dbg((CE_CONT,"cbq_close\n"));

	mutex_enter(&cbq_lock_g);
	cqp = (cbq_state_t *)q->q_ptr;

	switch (cqp->cbq_id) {
	case CBQ_DEVICE:
		qprocsoff(q);
		devp = &cqp->cbq.cbq_ds;

		if (!CBQ_ISCTL(devp->cbq_minor)) {
			for (mcqp = cbq_ill_head_g; mcqp;
					mcqp = CBQ_NEXT(mcqp)) {
				cbqp = GET_CBQ_MODULE(mcqp);

				cbqlock(WR(cbqp->cbq_q));

				if (!cbqp->use_esballoc)
					cbq_pktsched_destroy(cbqp);
				cbqp->cbq_enabled = DISABLE;
				cbq_clear_interface(cbqp);

				cbqunlock(WR(cbqp->cbq_q));
			}
		}
		devp->cbq_open = false;
		break;

	case CBQ_MODULE:
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(q));

		cbqp->cbq_enabled = DISABLE;
		untimeout(cbqp->cbq_timeout_id);
		cbq_clear_interface(cbqp);

		kmem_free(cbqp->cbq_class_tbl,
			sizeof (struct rm_class *) * CBQ_MAX_CLASSES);
		if (cbqp->cbq_stats_enabled) {
			if (cbqp->cbq_stats_timeout_id)
				untimeout(cbqp->cbq_stats_timeout_id);
			cbqp->cbq_stats_enabled = 0;
		}

		/*
		 * Remove cqp from the ill list.
		 */
		result = cbq_ill_delete_entry(cqp);
		if (!result)
			cmn_err(CE_WARN, "cbq_close: ill not deleted\n");

		/*
		 * Where done unlock and destroy lock.
		 */
		cbqunlock(WR(q));
		mutex_destroy(&cbqp->cbq_lock);
		break;
	}

	/*
	 * mi_close_comm frees the instance data structure.
	 */
	mi_close_comm(&cbq_head_g, q);
	q->q_ptr = (char *)NULL;
	mutex_exit(&cbq_lock_g);

	cbq0dbg((CE_CONT,"cbq_close: exit\n"));
	return (0);
}

static int
cbq_is_enabled(cbq_module_t *cbqp)
{
	int	result;

	cbqlock(WR(cbqp->cbq_q));
	result = cbqp->cbq_enabled;
	cbqunlock(WR(cbqp->cbq_q));
	return (result);
}

/*
 * static int
 * cbq_wput(queue_t *, mblk_t *mp) - This function handles streams message
 *	processing for both the cbq module and driver.
 *
 *	If we are the driver, then only M_IOCTL, M_IOCDATA and M_FLUSH
 *	messages are processed.  If we traveling down the module path,
 *	then IP wants to classify a packet or just send it on depending
 *	whether CBQ is enabled or not. All control messages such M_FLUSH
 *	and et al are just sent on always.
 *
 *	Returns:	0, for success.
 */

static int
cbq_wput(queue_t *q, mblk_t *mp)
{
	cbq_state_t		*cqp = (cbq_state_t *)q->q_ptr;
	cbq_module_t *cbqp;

	/*
	 * The streams module does not process any IOCTL's.  The streams
	 * driver is the only processing machine for IOCTL's.
	 */
	switch (cqp->cbq_id) {
	case CBQ_MODULE:
		putq(q, mp);
		qenable(q);
		break;


	case CBQ_DEVICE:
		switch (mp->b_datap->db_type) {
		case M_FLUSH:
			if (*mp->b_rptr & FLUSHW)
				flushq(q, FLUSHDATA);
			if (*mp->b_rptr & FLUSHR) {
				*mp->b_rptr &= ~FLUSHW;
				qreply(q, mp);
			} else
				freemsg(mp);
			break;

		case M_IOCTL:
			cbq_ioctl(q, mp);
			break;

		case M_IOCDATA:
			/* IOCTL continuation. */
			if (mi_copy_state(q, mp, (mblk_t **)NULL) == -1) {
				/*
				 * The copy operation failed.  mi_copy_state
				 * already cleaned up, so we can just return.
				 */
				return (0);
			}

			if (MI_COPY_DIRECTION(mp) == MI_COPY_IN)
				cbq_ioc(q, mp);
			else
				mi_copyout(q, mp);
			break;

		default:
			freemsg(mp);
			break;
		}
		break;

	default:
		cmn_err(CE_WARN, "cbq_wput: Illegal device or module\n");
		freemsg(mp);
		return (0);
	}

	return (0);
}

/*
 * static int
 * cbq_wsrv(queue_t *) - this function is only used by the cbq module
 *	to pass data from this module to the module downstream from us.
 *
 *	Returns:	0, on success.
 */

static int
cbq_wsrv(queue_t *q)
{
	mblk_t		*mp;
	cbq_state_t		*cqp = (cbq_state_t *)q->q_ptr;
	cbq_module_t *cbqp;

	switch (cqp->cbq_id) {
	case CBQ_MODULE:
		cbqp = GET_CBQ_MODULE(cqp);
		while ((mp = getq(q)) != NULL) {
			/*
		  	 * There are two cases for passing IP packets
			 * between IP and the data link layer. 
			 * DL_UNITDATA_REQ and M_DATA.  DL_UNITDATA if
			 * for non-fastpath capable interfaces,
			 * while M_DATA is fastpath capable.
			 */
			if (cbq_is_enabled(cbqp)) {
				switch (DB_TYPE(mp)) {
				case M_PROTO:
				case M_PCPROTO:
					/* Look for DL_UNITDATA_REQ */
					if (!cbq_proto(q, mp))
						return (0);
					break;

				case M_DATA:
					/* Fastpath */
					if (!cbqenque(q, mp, IS_FAST_PATH))
						goto cbq_wsrv_exit;
					break;

				default:
					if (canputnext(q))
						putnext(q, mp);
					else {
						putbq(q, mp);
						goto cbq_wsrv_exit;
					}
					break;
				}
			} else if (canputnext(q)) {
				putnext(q, mp);
			} else {
				putbq(q, mp);
				goto cbq_wsrv_exit;
			}
		}
		break;

	case CBQ_DEVICE:
		break;
	}

cbq_wsrv_exit:
	return (0);
}

/*
 * Static int
 * cbq_rput(queue_t *, mblk_t *) - This functions passes the data upstream
 *	from the cbq Module or driver.
 *
 *	Returns:	0, for success.
 */

static int
cbq_rput(queue_t *q, mblk_t *mp)
{
	putq(q, mp);
	return (0);
}

static int
cbq_rsrv(queue_t *q)
{
	mblk_t		*hp, *mp;
	struct iocblk	*iocp;
	cbq_state_t		*cqp;
	cbq_module_t	*cbqp;
	int		dtype, just_free_it = 0;

	cqp = (cbq_state_t *)WR(q)->q_ptr;

	while ((mp = getq(q)) != NULL) {
		dtype = DB_TYPE(mp);
		switch (dtype) {
		case M_IOCACK:
		case M_IOCNAK:
			if (cqp->cbq_id == CBQ_MODULE) {
				cbqp = GET_CBQ_MODULE(cqp);
				cbqlock(WR(q));
				/*
				 * Does the driver implement the fastpath
				 * IOCTL.
				 */
				iocp = (struct iocblk *)mp->b_rptr;
				switch (iocp->ioc_cmd) {
				case DL_IOC_HDR_INFO:
					if (dtype == M_IOCACK) {
						hp = mp->b_cont->b_cont;
						cbqp->mac_addr_sz = MBLKL(hp);
					} else
						cbqp->mac_addr_sz = 0;
					break;
				case DL_IOC_PKTSCHED_INIT:
					if (dtype == M_IOCACK)
						cbqp->is_pktsched_inited = 1;
					else
						cbqp->is_pktsched_inited = -1;
					cv_broadcast(&cbqp->cv);
					just_free_it = 1;
					break;
				case DL_IOC_PKTSCHED_TEARDOWN:
					if (dtype == M_IOCACK)
						cbqp->is_pktsched_inited = 0;
					else
						cbqp->is_pktsched_destroyed
							= -1;
					cv_broadcast(&cbqp->cv);
					just_free_it = 1;
					break;
				}
				cbqunlock(WR(q));
			}
			break;
		default:
			break;
		}

		if (just_free_it) {
			freemsg(mp);
			just_free_it = 0;
		} else {
			if (canputnext(q))
				putnext(q, mp);
			else {
				putbq(q, mp);
				return (0);
			}
		}
	}

	return (0);
}

/*
 * static void
 * cbq_ioctl(queue_t *, mblk_t *) - Start the processing of CBQ device
 *	ioctls.  The actual processing is done in cbq_ioc.
 *
 *	Returns:	NONE.
 */

static void
cbq_ioctl(queue_t *q, mblk_t *mp)
{
	struct copyreq		*cq;
	cbq_state_t			*cqp = (cbq_state_t *)q->q_ptr;
	struct iocblk		*iocp = (struct iocblk *)mp->b_rptr;
	int			error = 0, copy_in_size = 0;
	caddr_t			uaddr = (caddr_t)NULL;

	cbq0dbg((CE_CONT,"cbq_ioctl\n"));

	if (!mp->b_cont) {
		error = EPROTO;
		goto cbq_ioctl_done;
	}

	/*
	 * All of the ioctl's execpt for CBQ_DUMP_ILL_LIST require
	 * COPYIN to happen.  The actual ioctl is processed in cbq_ioc
	 * once the ioctl data is sent down to the CBQ device via the
	 * M_IOCDATA message.
	 */

	switch (iocp->ioc_cmd) {
	case CBQ_GETSTATS:
		copy_in_size = sizeof (struct cbq_getstats);
		break;

	case CBQ_ENABLE:
	case CBQ_DISABLE:
	case CBQ_ENABLE_STATS:
	case CBQ_DISABLE_STATS:
	case CBQ_CLEAR_HIERARCHY:
		copy_in_size = sizeof (struct cbq_interface);
		break;

	case CBQ_ADD_FILTER:
		copy_in_size = sizeof (struct cbq_add_filter);
		break;

	case CBQ_DEL_FILTER:
		copy_in_size = sizeof (struct cbq_delete_filter);
		break;

	case CBQ_ADD_CLASS:
		copy_in_size = sizeof (struct cbq_add_class);
		break;

	case CBQ_DEL_CLASS:
		copy_in_size = sizeof (struct cbq_delete_class);
		break;
#ifdef CBQPERF
	case CBQ_GETNPERF:
		copy_in_size = sizeof(struct cbq_getnperf);
		break;
	case CBQ_GETPERF:
		copy_in_size = sizeof(struct cbq_getperf);
		break;
#endif

	default:
		error = EINVAL;
		break;
	}

cbq_ioctl_done:
	if (error) {
		if (mp->b_cont) {
			freemsg(mp->b_cont);
			mp->b_cont = NULL;
		}
		mp->b_datap->db_type = M_IOCNAK;
		iocp->ioc_count = 0;
		iocp->ioc_error = error;
		qreply(q, mp);
	} else
		mi_copyin(q, mp, nilp(char), copy_in_size);

	cbq0dbg((CE_CONT,"cbq_ioctl: exit\n"));
	return;
}

/*
 * static void
 * cbq_ioc(queue_t *, mblk_t *) - this functions performs ioctl's on the
 *	cbq device.  This function is called as result of data being
 *	copied in form user address space.  The stream copies the data in
 *	and then sends a M_IOCDATA message downstream.
 *
 *	Returns:	NONE
 */

static void
cbq_ioc(queue_t *q, mblk_t *mp)
{
	caddr_t				addr;
	struct cbq_add_filter		*afp;
	struct cbq_delete_filter	*dfp;
	struct cbq_add_class		*acp;
	struct cbq_delete_class		*dcp;
#ifdef CBQPERF
	struct cbq_getnperf		*getnpfp;
	struct cbq_getperf		*getpfp;
#endif
	mblk_t				*bp;
	struct iocblk			*iocp = (struct iocblk *)mp->b_rptr;
	int				error = 0;

	cbq0dbg((CE_CONT,"cbq_ioc\n"));

	if (!(bp = mp->b_cont) || !(bp = bp->b_cont)) {
		error = EPROTO;
		goto cbq_ioc_done;
	}

	/* Get pointer to ioctl data. */
	addr = (char *)bp->b_rptr;
	switch (iocp->ioc_cmd) {
	case CBQ_GETSTATS:
		error = cbq_getstats((struct cbq_getstats *)addr);
		break;

	case CBQ_ENABLE_STATS:
	case CBQ_DISABLE_STATS:
		break;

	case CBQ_ENABLE:
		error = cbq_set_enable((struct cbq_interface *)addr, ENABLE);
		break;

	case CBQ_DISABLE:
		error = cbq_set_enable((struct cbq_interface *)addr, DISABLE);
		break;

	case CBQ_ADD_FILTER:
		afp = (struct cbq_add_filter *)addr;
		if ((error = cbq_add_filter(afp)) == 0)
			goto cbq_ioc_copyout;
		break;

	case CBQ_DEL_FILTER:
		dfp = (struct cbq_delete_filter *)addr;
		error = cbq_delete_filter(dfp);
		break;

	case CBQ_ADD_CLASS:
		acp = (struct cbq_add_class *)addr;

		if ((error = cbq_add_class(acp)) == 0)
			goto cbq_ioc_copyout;
		break;

	case CBQ_DEL_CLASS:
		dcp = (struct cbq_delete_class *)addr;
		error = cbq_delete_class(dcp);
		break;

	case CBQ_CLEAR_HIERARCHY:
		error = cbq_clear_hierarchy((struct cbq_interface *)addr);
		break;
#ifdef CBQPERF
	case CBQ_GETNPERF:
		getnpfp = (struct cbq_getnperf *)addr;
		if ((error =  cbq_getnperf(getnpfp)) == 0)
			goto cbq_ioc_copyout;
		break;
	case CBQ_GETPERF:
		getpfp = (struct cbq_getperf *)addr;
		if ((error =  cbq_getperf(getpfp)) == 0)
			goto cbq_ioc_copyout;
		break;
#endif
	default:
		error = EINVAL;
		break;
	}

cbq_ioc_done:
	mi_copy_done(q, mp, error);
	cbq0dbg((CE_CONT,"cbq_ioc(done)\n"));
	return;

cbq_ioc_copyout:
	mi_copyout(q, mp);
	cbq0dbg((CE_CONT,"cbq_ioc(copyout)\n"));
	return;
}

#ifdef CBQPERF
int
cbq_initperf(queue_t *q)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE((cbq_state_t *)q->q_ptr);

	if (cbqp->perf)
		cbq_destroyperf(q);

	if ((cbqp->perf = kmem_zalloc(sizeof (struct cbq_perf) * N_CBQ_PERF,
			KM_SLEEP)) == NULL) {
		cmn_err(CE_CONT, "cbq_initperf: kmem_zalloc failed\n");
		return (1);
	}

	cbqp->nperf = 0;
	return(0);
}

int
cbq_destroyperf(queue_t *q)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE((cbq_state_t *)q->q_ptr);

	if (cbqp->perf)
		kmem_free(cbqp->perf, sizeof (struct cbq_perf) * N_CBQ_PERF);
	cbqp->nperf = 0;
	return (0);
}

static int
cbq_getnperf(struct cbq_getnperf *getnpfp)
{
	int		error = 0;
	cbq_state_t	*cqp;
	cbq_module_t	*cbqp;
	char		*ifacename;
	int		ifacelen;

	/*
	 * Match interface name to interface state.
	 */
	ifacename = getnpfp->iface.cbq_ifacename;
	ifacelen = getnpfp->iface.cbq_ifacelen;

	mutex_enter(&cbq_lock_g);
	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) == NULL) {
		error = EINVAL;
		mutex_exit(&cbq_lock_g);
		return (error);
	}
	mutex_exit(&cbq_lock_g);

	cbqp = GET_CBQ_MODULE(cqp);
	cbqlock(WR(cbqp->cbq_q));
	getnpfp->nrecords = cbqp->nperf;
	cbqunlock(WR(cbqp->cbq_q));
	return (error);
}

static int
cbq_getperf(struct cbq_getperf *getpfp)
{
	int		error = 0;
	cbq_state_t	*cqp;
	cbq_module_t	*cbqp;
	char		*ifacename;
	int		ifacelen;

	/*
	 * Match interface name to interface state.
	 */
	ifacename = getpfp->iface.cbq_ifacename;
	ifacelen = getpfp->iface.cbq_ifacelen;
									    
	mutex_enter(&cbq_lock_g);
	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) == NULL) {
		error = EINVAL;
		mutex_exit(&cbq_lock_g);
		return (error);
	}
	mutex_exit(&cbq_lock_g);

	cbqp = GET_CBQ_MODULE(cqp);
	cbqlock(WR(cbqp->cbq_q));
	if (getpfp->id >= cbqp->nperf)
		error = EINVAL;
	else 
		getpfp->perf = cbqp->perf[getpfp->id];
	cbqunlock(WR(cbqp->cbq_q));
	return (error);
}

void
cbq_perf(queue_t *q, int class_id, hrtime_t schedtime, hrtime_t endtime)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE((cbq_state_t *)q->q_ptr);

	cbqp->perf[cbqp->nperf].class_id = class_id;
	cbqp->perf[cbqp->nperf].schedtime = schedtime;
	cbqp->perf[cbqp->nperf].departtime = endtime;

	cbqp->nperf++;
	if (cbqp->nperf >= N_CBQ_PERF)
		cbqp->nperf = 0;
}

#endif

static int
cbq_proto(queue_t *q, mblk_t *mp)
{
	union DL_primitives	*dlp;
	u_long			prim;
	mblk_t			*nmp;

	dlp = (union DL_primitives *)mp->b_rptr;
	prim = dlp->dl_primitive;

	if ((prim == DL_UNITDATA_REQ) && (((nmp = mp->b_cont) != NULL) &&
			(DB_TYPE(nmp) == M_DATA))) {
		return (cbqenque(q, mp, IS_NOT_FAST_PATH));
	} else if (canputnext(q)) {
		putnext(q, mp);
		return (1);
	}
	
	putbq(q, mp);
	return (0);
}

/*
 * static int
 * cbq_set_enable(struct cbq_enable *ep) - this function processed the
 *	ioctl request to enable class based queueing.  It searches the list
 *	of interfaces for the specified interface and then enables CBQ on
 *	that interface.
 *
 *	Returns:	0, for no error.
 *			ENXIO, for specified inteface not found.
 */

static int
cbq_set_enable(struct cbq_interface *ep, int enable)
{
	int 		error = 0;
	cbq_state_t		*cqp;
	cbq_module_t	*cbqp;
	char		*ifacename;
	int		ifacelen;

	cbq0dbg((CE_CONT,"cbq_set_enable\n"));

	ifacename = ep->cbq_ifacename;
	ifacelen = ep->cbq_ifacelen;

	mutex_enter(&cbq_lock_g);
	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen))) {
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(cbqp->cbq_q));

		switch (enable) {
		case ENABLE:
			if (!cbqp->ifnp.root_ || !cbqp->ifnp.default_ ||
					!cbqp->ifnp.ctl_) {
				cbqp->cbq_enabled = DISABLE;
				if (!cbqp->ifnp.root_)
					cmn_err(CE_WARN,
						"No Root Class for %s\n",
						ifacename);
				if (!cbqp->ifnp.default_)
					cmn_err(CE_WARN,
						"No Default Class for %s\n",
						ifacename);
				if (!cbqp->ifnp.ctl_)
					cmn_err(CE_WARN,
						"No Control Class for %s\n",
						ifacename);
				error = ENXIO;
			} else if (cbqp->ifnp.root_ == cbqp->ifnp.default_) {
				cmn_err(CE_CONT, "Invalid default class");
				error = ENXIO;
			} else {
				if (!cbq_pktsched_init(cbqp)) {
					cbqp->use_esballoc++;
					cbqp->ifnp.maxqueued_ = RM_MAXQUEUED;
				} 
				cbqp->cbq_enabled = enable;
			}
			break;

		case DISABLE:
			if (!cbqp->use_esballoc)
				cbq_pktsched_destroy(cbqp);
			cbqp->cbq_enabled = enable;
			break;
		}
		cbqunlock(WR(cbqp->cbq_q));
		cbq_satisfied(cbqp);
	} else
		error = ENXIO;
	mutex_exit(&cbq_lock_g);

	cbq0dbg((CE_CONT,"cbq_set_enable: exit\n"));
	return (error);
}

/*
 * void
 * cbq_satisfied(cbqp) - Check CBQ to see how link-sharing is performing.
 *
 *	Returns: NONE
 */

static void
cbq_satisfied(cbq_module_t *cbqp)
{
#ifdef USE_HRTIME
	hrtime_t	now;
#else
	struct timeval	now;
#endif
#ifdef notyet
	cbqlock(WR(cbqp->cbq_q));
	if (cbqp->cbq_enabled) {
		RM_GETTIME(now);
		rmc_tl_satisfied(&cbqp->ifnp, &now);
		cbqp->cbq_state_tl_timeout_id = qtimeout(cbqp->cbq_q,
			cbq_satisfied, (caddr_t)cbqp, CBQ_LS_TIMEOUT);
	} else
		cbqp->cbq_state_tl_timeout_id = 0;
	cbqunlock(WR(cbqp->cbq_q));
#endif
}

static int
cbq_getstats(struct cbq_getstats *gs)
{
	int		error = 0;
	cbq_state_t		*cqp;
	cbq_module_t	*cbqp;
	char		*ifacename;
	int		ifacelen;

	ifacename = gs->iface.cbq_ifacename;
	ifacelen = gs->iface.cbq_ifacelen;

	mutex_enter(&cbq_lock_g);
	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) == NULL) {
		error = EINVAL;
		mutex_exit(&cbq_lock_g);
		return (error);
	}
	mutex_exit(&cbq_lock_g);

	cbqp = GET_CBQ_MODULE(cqp);
	cbqlock(WR(cbqp->cbq_q));
	error = _cbq_getstats(gs, cbqp);
	cbqunlock(WR(cbqp->cbq_q));

	return (error);
}

static int
_cbq_getstats(struct cbq_getstats *gs, cbq_module_t *cbqp)
{
	int		nclasses = gs->nclasses, i, error = 0;
	struct rm_class *cl;
	caddr_t		uaddr;

	/* Get the User Address Space pointer. */
	uaddr = (caddr_t)gs->stats;

	/*
	 * Copyout the root class.
	 */
	cl = cbqp->ifnp.root_;
	cl->stats_.depth = cl->depth_;
	if (copyout((caddr_t)&cl->stats_, uaddr, sizeof (class_stats_t))) {
		error = EFAULT;
		goto _cbq_getstats_exit;
	}

	/*
	 * Copyout the default class.
	 */
	cl = cbqp->ifnp.default_;
	cl->stats_.depth = cl->depth_;
	uaddr += sizeof (class_stats_t);
	if (copyout((caddr_t)&cl->stats_, uaddr, sizeof (class_stats_t))) {
		error = EFAULT;
		goto _cbq_getstats_exit;
	}

	/*
	 * Copyout the control class.
	 */
	cl = cbqp->ifnp.ctl_;
	cl->stats_.depth = cl->depth_;
	uaddr += sizeof (class_stats_t);
	if (copyout((caddr_t)&cl->stats_, uaddr, sizeof (class_stats_t))) {
		error = EFAULT;
		goto _cbq_getstats_exit;
	}
	nclasses -= 3;

	/*
	 * Get the rest of the classes.
	 */
	for (i = 0; (i < CBQ_MAX_CLASSES) && nclasses; i++) {
		if (cbqp->cbq_class_tbl[i] != NULL) {
			cl = cbqp->cbq_class_tbl[i];
			cl->stats_.depth = cl->depth_;
			uaddr += sizeof (class_stats_t);
			if (copyout((caddr_t)&cl->stats_, uaddr,
					sizeof (class_stats_t))) {
				error = EFAULT;
				goto _cbq_getstats_exit;
			}
			nclasses--;
		}
	}

_cbq_getstats_exit:
	return (error);
}

static int
cbq_check_filter_handle(k_filter_cb_t *flist, u_int handle)
{
	k_filter_cb_t	*kfp;

	for (kfp = flist; kfp != NULL; kfp = kfp->next) {
		if (kfp->handle == handle)
			return (1);
	}

	return (0);
}

static int
cbq_add_filter(struct cbq_add_filter *afp)
{
	int			error = 0;
	u_int			fhi;
	char			*ifacename;
	int			i, ifacelen;
	cbq_state_t			*cqp;
	cbq_module_t		*cbqp;
	k_filter_cb_t		*kfcbp, *kfp, *cl_filt_p;
	struct rm_class		*classp;

	u_int			hindex;
	hrtime_t		hrtime;

	cbq0dbg((CE_CONT,"cbq_add_filter\n"));

	ifacename = afp->cbq_iface.cbq_ifacename;
	ifacelen = afp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(cbqp->cbq_q));

		if (afp->cbq_filter.daddr == 0) {
			error = EINVAL;
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_filter_exit;
		}

		/* Get the pointer to class. */
		classp = cbq_clh_to_clp(cbqp, afp->cbq_class_handle, &error);
		if (error) {
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_filter_exit;
		}

		if ((kfcbp = kmem_zalloc(sizeof (k_filter_cb_t),
				KM_SLEEP)) == NULL) {
			error = ENOMEM;
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_filter_exit;
		}

		kfcbp->filter = afp->cbq_filter;
		kfcbp->tbfilter.enable = 0;

		/*
		 * Generate Handle for filter.
		 */
		hindex = CBQ_GET_HASH_INDEX(kfcbp->filter.daddr);
cbq_get_handle:
		hrtime = gethrtime();
		kfcbp->handle = (hindex << 24) |
			(u_int)((hrtime & 0x00ffffff));
		if (cbq_check_filter_handle(cbqp->filters[hindex],
				kfcbp->handle))
			goto cbq_get_handle;

		kfcbp->clp = classp;

		/*
		 * Add filter to front of filter list.
		 */
		kfcbp->prev = NULL;
		kfcbp->next = cbqp->filters[hindex];
		if (cbqp->filters[hindex])
			cbqp->filters[hindex]->prev = kfcbp;
		cbqp->filters[hindex] = kfcbp;
		afp->cbq_filter_handle = kfcbp->handle;
		cbqunlock(WR(cbqp->cbq_q));
	} else
		error = ENXIO;

cbq_add_filter_exit:
	cbq0dbg((CE_CONT,"cbq_add_filter: exit\n"));
	return (error);
}

static int
cbq_delete_filter(struct cbq_delete_filter *dfp)
{
	int			error = 0;
	char			*ifacename;
	int			ifacelen;
	struct rm_class		*classp;
	cbq_state_t			*cqp;
	u_long			f_handle;
	cbq_module_t		*cbqp;
	k_filter_cb_t		*kfp;

	cbq0dbg((CE_CONT,"cbq_delete_filter\n"));

	ifacename = dfp->cbq_iface.cbq_ifacename;
	ifacelen = dfp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(cbqp->cbq_q));

		kfp = cbq_filth_to_filtp(cbqp, dfp->cbq_filter_handle, &error);
		if (error) {
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_delete_filter_exit;
		}

		cbq_dealloc_filter(cbqp, kfp);
		cbqunlock(WR(cbqp->cbq_q));
	} else
		error = ENXIO;

cbq_delete_filter_exit:
	cbq0dbg((CE_CONT,"cbq_delete_filter\n"));
	return (error);
}

static int
cbq_add_class(struct cbq_add_class *acp)
{
	int			error = 0;
	char			*ifacename;
	int			ifacelen;
	struct rm_class		*classp, *borrow, *parent;
	cbq_state_t			*cqp;
	cbq_module_t		*cbqp;
	u_long			chandle;

	cbq0dbg((CE_CONT,"cbq_add_class\n"));

	ifacename = acp->cbq_iface.cbq_ifacename;
	ifacelen = acp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(cbqp->cbq_q));

		/* Get pointers to parent and borrow classes.  */
		parent = cbq_clh_to_clp(cbqp,
			acp->cbq_class.parent_class_handle, &error);
		if (error)  {
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_class_exit;
		}

		borrow = cbq_clh_to_clp(cbqp,
			acp->cbq_class.borrow_class_handle, &error);
		if (error) {
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_class_exit;
		}

		/*
		 * A class must borrow from it's parent or it can not
		 * borrow at all.  Hence, borrow can be null.
		 */
		if ((borrow != parent)  && (borrow != NULL)) {
			cmn_err(CE_CONT,
				"cbq_add_class: borrow class != parent\n");
			error = EINVAL;
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_class_exit;
		}

		/* Create class.  */
		if (!parent) {
			if (cbqp->ifnp.root_) {
				error = EINVAL;
				cbqunlock(WR(cbqp->cbq_q));
				goto cbq_add_class_exit;
			}

			chandle = ROOT_CLASS_HANDLE;
		} else if (acp->cbq_class.defaultclass) {
			if (cbqp->ifnp.default_) {
				error = EINVAL;
				cbqunlock(WR(cbqp->cbq_q));
				goto cbq_add_class_exit;
			}

			chandle = DEFAULT_CLASS_HANDLE;
		} else if (acp->cbq_class.ctlclass) {
			if (cbqp->ifnp.ctl_) {
				error = EINVAL;
				cbqunlock(WR(cbqp->cbq_q));
				goto cbq_add_class_exit;
			}
			chandle = CTL_CLASS_HANDLE;
		} else {
			chandle = cbq_alloc_class_handle(cbqp, &error);
			if (error) {
				cbqunlock(WR(cbqp->cbq_q));
				goto cbq_add_class_exit;
			}
		}

		if ((classp = cbq_class_create(cbqp, acp, chandle,
				parent, borrow)) == NULL) {
			cbq_free_class_handle(cbqp, chandle);
			error = ENOMEM;
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_add_class_exit;
		}

		/* Return handle to user space. */
		classp->stats_.handle = chandle;
		classp->stats_.depth = classp->depth_;
		acp->cbq_class_handle = chandle;
		cbqunlock(WR(cbqp->cbq_q));
	} else
		error = ENXIO;

cbq_add_class_exit:
	cbq0dbg((CE_CONT,"cbq_add_class: exit\n"));
	return (error);
}

static int
cbq_delete_class(struct cbq_delete_class *dcp)
{
	int			i, error = 0;
	char			*ifacename;
	int			ifacelen;
	struct rm_class		*classp;
	cbq_state_t			*cqp;
	cbq_module_t		*cbqp;
	k_filter_cb_t		*kfp, *tfp;

	cbq0dbg((CE_CONT,"cbq_delete_class\n"));

	ifacename = dcp->cbq_iface.cbq_ifacename;
	ifacelen = dcp->cbq_iface.cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(cbqp->cbq_q));

		classp = cbq_clh_to_clp(cbqp, dcp->cbq_class_handle, &error);
		if (error || !classp) {
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_del_class_exit;
		}

		/*
		 * If a filter has a reference to this class delete
		 * the filter before deleting the class.
		 */
		for (i = 0; i < CBQ_FILTER_TBL_SZ; i++) {
			for (kfp = cbqp->filters[i]; kfp; kfp = tfp) {
				tfp = kfp->next;
				if (kfp->clp == classp)
					cbq_dealloc_filter(cbqp, kfp);
			}
		}

		/*
		 * If we are a parent class, then return an error.
		 */
		if (is_a_parent_class(cbqp, classp)) {
			error = EINVAL;
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_del_class_exit;
		}

		error = cbq_class_destroy(cbqp, classp);
		if (error) {
			cbqunlock(WR(cbqp->cbq_q));
			goto cbq_del_class_exit;
		}

		cbq_free_class_handle(cbqp, dcp->cbq_class_handle);
		cbqunlock(WR(cbqp->cbq_q));
	} else
		error = ENXIO;

cbq_del_class_exit:
	cbq0dbg((CE_CONT,"cbq_delete_class\n"));
	return (error);
}

static int
cbq_clear_hierarchy(struct cbq_interface *ifp)
{
	int		ifacelen, error = 0;
	char		*ifacename;
	cbq_state_t	*cqp;
	cbq_module_t	*cbqp;

	cbq0dbg((CE_CONT,"cbq_clear_hierarchy\n"));

	ifacename = ifp->cbq_ifacename;
	ifacelen = ifp->cbq_ifacelen;

	if ((cqp = cbq_ill_lookup_on_name(ifacename, ifacelen)) != NULL) {
		cbqp = GET_CBQ_MODULE(cqp);
		cbqlock(WR(cbqp->cbq_q));

		if (!cbqp->use_esballoc)
			cbq_pktsched_destroy(cbqp);

		cbqp->cbq_enabled = DISABLE;
		error = cbq_clear_interface(cbqp);

		cbqunlock(WR(cbqp->cbq_q));

	} else
		error = ENXIO;

cbq_clear_hierarchy_exit:
	cbq0dbg((CE_CONT,"cbq_clear_hierarchy: exit\n"));
	return (error);
}

static int
cbq_clear_interface(cbq_module_t *cbqp)
{
	int		again, i;
	k_filter_cb_t	*kfp, *tfp;
	struct rm_class	*cl;

	/*
	 * Free the filters for this interface.
	 */
	for (i = 0; i < CBQ_FILTER_TBL_SZ; i++) {
		for (kfp = cbqp->filters[i]; kfp; kfp = tfp) {
			tfp = kfp->next;
			cbq_dealloc_filter(cbqp, kfp);
		}
	}

	/*
	 * Clear out the classes now.
	 */
do_it_again:
	again = 0;
	for (i = 0; i < CBQ_MAX_CLASSES; i++) {
		if ((cl = cbqp->cbq_class_tbl[i]) == NULL)
			continue;
		if (is_a_parent_class(cbqp, cl)) {
			again++;
			continue;
		}

		cbq_class_destroy(cbqp, cl);
		cbqp->cbq_class_tbl[i] = NULL;
	}
	if (again)
		goto do_it_again;

	if (cbqp->ifnp.ctl_) {
		cbq_class_destroy(cbqp, cbqp->ifnp.ctl_);
		cbqp->ifnp.ctl_ = NULL;
	}

	if (cbqp->ifnp.default_) {
		cbq_class_destroy(cbqp, cbqp->ifnp.default_);
		cbqp->ifnp.default_ = NULL;
	}

	if (cbqp->ifnp.root_) {
		cbq_class_destroy(cbqp, cbqp->ifnp.root_);
		cbqp->ifnp.root_ = NULL;
	}
	return (0);
}

static int
is_a_parent_class(cbq_module_t *cbqp, struct rm_class *cl)
{
	int	i;

	if (!cl)
		return (0);

	for (i = 0; i < CBQ_MAX_CLASSES; i++) {
		if (cbqp->cbq_class_tbl[i]) {
			if (cbqp->cbq_class_tbl[i]->parent_ == cl)
				return (1);
		}
	}

	return (0);
}

static void
cbq_dealloc_filter(cbq_module_t *cbqp, k_filter_cb_t *kfp)
{
	u_int	hindex;

	if (!kfp || !cbqp)
		return;

	/* Delete from the filter list. */
	hindex = CBQ_GET_HINDEX(kfp->handle);

	if (kfp->prev && kfp->next) {
		kfp->prev->next = kfp->next;
		kfp->next->prev = kfp->prev;
	} else if (kfp->prev && !kfp->next) {
		kfp->prev->next = NULL;
	} else if (kfp->next && !kfp->prev) {
		kfp->next->prev = NULL;
		cbqp->filters[hindex] = kfp->next;
	} else
		cbqp->filters[hindex] = NULL;

	kfp->next = kfp->prev = NULL;
	kmem_free((void *)kfp, sizeof (k_filter_cb_t));
}

static k_filter_cb_t *
cbq_filth_to_filtp(cbq_module_t *cbqp, u_long handle, int *error)
{
	k_filter_cb_t 	*kfp;
	u_int		hindex;

	hindex = CBQ_GET_HINDEX(handle);
	if (cbqp->filters[hindex]) {
		for (kfp = cbqp->filters[hindex]; kfp; kfp = kfp->next) {
			if (kfp->handle == handle)
				return (kfp);
		}
	}

	*error = EINVAL;
	return ((k_filter_cb_t *)NULL);
}

static u_long
cbq_alloc_class_handle(cbq_module_t *cbqp, int *error)
{
	int	i;
	u_long	handle = 0;

	for (i = 0; i < CBQ_MAX_CLASSES; i++) {
		if (!cbqp->cbq_class_tbl[i])
			break;
	}

	if (i >= CBQ_MAX_CLASSES)
		*error = ENOSPC;

	handle = (u_long)i;
	return (handle);
}

static void
cbq_free_class_handle(cbq_module_t *cbqp, u_long handle)
{
	switch (handle) {
	case ROOT_CLASS_HANDLE:
	case DEFAULT_CLASS_HANDLE:
	case CTL_CLASS_HANDLE:
		break;

	default:
		if (handle >= CBQ_MAX_CLASSES)
			return;
		cbqp->cbq_class_tbl[handle] = NULL;
	}
}

static struct rm_class *
cbq_clh_to_clp(cbq_module_t *cbqp, u_long handle, int *error)
{
	switch (handle) {
	case NULL_CLASS_HANDLE:
		return ((struct rm_class *)NULL);

	case ROOT_CLASS_HANDLE:
		if (!cbqp->ifnp.root_)
			*error = EINVAL;
		return (cbqp->ifnp.root_);

	case DEFAULT_CLASS_HANDLE:
		if (!cbqp->ifnp.default_)
			*error = EINVAL;
		return (cbqp->ifnp.default_);

	case CTL_CLASS_HANDLE:
		if (!cbqp->ifnp.ctl_)
			*error = EINVAL;
		return (cbqp->ifnp.ctl_);

	default:
		if (handle >= CBQ_MAX_CLASSES)
			*error = ENXIO;
		else
			return (cbqp->cbq_class_tbl[handle]);
	}
}

/*
 * static int
 * cbq_ill_init(queue_t *, cbq_state_t *) - this functions is called when CBQ is
 *	pushed onto the stream to the network interface.  This routine mainly
 *	gets the name of the network interface below it and adds it to a list
 *	network interfaces that CBQ has pushed onto.
 *
 *	Returns:	ENXIO, if no such interface exists.
 *			0, if no error occurred.
 */

static int
cbq_ill_init(queue_t *q, cbq_state_t *cqp)
{
	queue_t	*tq;
	char	*cp;
	int	count;

	cbq0dbg((CE_CONT,"cbq_ill_init\n"));

	/*
	 * Don't look! We walk down the stream to get the name of the
	 * driver so we can constrcut a the ILL name for this module.
	 */
	for (tq = q; (tq && tq->q_next); tq = tq->q_next)
		;
	cp = tq->q_qinfo->qi_minfo->mi_idname;
	if (!cp || !*cp) {
		cbq0dbg((CE_CONT,"cbq_ill_init: FAIL\n"));
		return (ENXIO);
	}

	/*
	 * Contruct a name from our devices 'q_qinfo->qi_minfo->mi_idname'
	 * and the lowest 'unit number' that is unused.  We assume here
	 * that devices are opened in ascending order.
	 */
	count = -1;
	do {
		if (++count > 1000) {
			cbq0dbg((CE_CONT,"cbq_ill_init: FAIL\n"));
			return (ENXIO);
		}

		sprintf(CBQ_ILL_NAME(cqp), "%s%d", cp, count);
		CBQ_ILL_NAMELEN(cqp) = strlen(CBQ_ILL_NAME(cqp));

	} while (cbq_ill_lookup_on_name(CBQ_ILL_NAME(cqp),
			CBQ_ILL_NAMELEN(cqp)));

	CBQ_ILL_PPA(cqp) = count;
	CBQ_NEXT(cqp) = NULL;

	/*
	 * Append to the list of cbq_ill's.
	 */
	if (cbq_ill_head_g) {
		cbq_state_t	*illp;

		illp = cbq_ill_head_g;
		while (illp->cbq_next)
			illp = illp->cbq_next;
		illp->cbq_next = cqp;
	} else
		cbq_ill_head_g = cqp;

	cbq0dbg((CE_CONT,"cbq_ill_init: SUCCESS\n"));
	return (0);
}

/*
 * cbq_state_t *
 * cbq_ill_lookup_on_name(char *, int) - return the pointer to the cbq_state_t
 *	structure given given by name and namelen.
 *
 *	Returns:	NULL, if unable to find the interface.
 *			pointer to cbq_state_t structure.
 */

cbq_state_t *
cbq_ill_lookup_on_name(char *name, int namelen)
{
	cbq_state_t	*cqp = NULL;
	int	i;

	if (namelen == 0)
		return (cqp);

	for (cqp = cbq_ill_head_g; cqp; cqp = cqp->cbq_next) {
		if ((CBQ_ILL_NAMELEN(cqp) == namelen) &&
				(strncmp(CBQ_ILL_NAME(cqp), name,
				namelen) == 0)) 
			break;
	}

	return (cqp);
}

/*
 * static int
 * cbq_ill_delete_entry(cbq_state_t *dcqp)
 *
 * Deletes 'dcpq' from cbq_ill_head_g list, relinking correctly the list
 * itself.
 *
 * Returns:	0: Element not found.
 *		1: Element deleted.
 */

static int
cbq_ill_delete_entry(cbq_state_t *dcqp)
{
	cbq_state_t	*pcqp = NULL, *cqp = NULL;
	int	i;

	/*
	 * Find the entry in the list of 'ills'.
	 */
	for (cqp = cbq_ill_head_g; cqp; cqp = cqp->cbq_next) {
		if (cqp == dcqp)
			break;
		else
			pcqp = cqp;
	}

	if (!cqp)
		return (0);

	/*
	 * Update its neighbors.
	 */
	if (!pcqp)
		cbq_ill_head_g = CBQ_NEXT(dcqp);
	else
		pcqp->cbq_next = CBQ_NEXT(dcqp);
	CBQ_NEXT(dcqp) = NULL;

	/* Return OK condition.  */
	return (1);
}

/*
 * void
 * cbqenque(queue_t *q, mblk_t *mp, int fastpath) - Queue data packets.
 *
 *	Assumptions:	State lock help by this thread when entering.
 *	Returns:	1, if packet a packet was queued downstream
 *			0, if not
 */

int
cbqenque(queue_t *q, mblk_t *mp, int fastpath)
{
	int	err;

	cbqlock(WR(q));

	/*
	 * If this message appears all bets are off
	 */
	if (mp->b_next) {
		cmn_err(CE_WARN, "cbqenque: n_next != NULL\n");
		mp->b_next = 0;
	}

	cbq_class_queue(WR(q), mp, fastpath);

	/*
 	 * XXX - This is not quite correct we shouldn't need to
 	 * bootstrap the first sending of a packet like this
 	 */
	err = cbq_deque(WR(q));
	cbqunlock(WR(q));
	
	return (err);
}

/*
 * void
 * cbqstart(queue_t *q, mblk_t *mp, *cbq_module_t *cbqp)
 *
 *	As we are queueing messages ourselves and only passing messages down
 *	to the mac driver one message at a time we need to know when the
 *	previous packet has been transmitted.  We the the mac completion
 *	interrupt by using esballoc to allocate a new header.  Esballoc
 *	allows the caller to supply its own free routine.  So, whenever the
 *	free routine is called we know the packet has been put out on the
 *	wire.
 *
 *	Returns:	NONE
 */

void
cbqstart(queue_t *q, mblk_t *mp, cbq_module_t *cbqp)
{
	mblk_t		*bp, *nmp;

	/*
	 * If the device below does not implement pktsched hooks
	 * resort to using esballoc.
	 */
	if (cbqp->use_esballoc) {
		if ((nmp = esballoc(mp->b_rptr, mp->b_wptr - mp->b_rptr,
				BPRI_HI, &cbqp->cbq_retn)) == 0) {
			cmn_err(CE_WARN, "cbqstart: esballoc failed\n");
			freemsg(mp);
			return;
		}
		cbqp->cbq_freeit[cbqp->cbqfi] = mp;
		cbqp->cbqfi = (cbqp->cbqfi + 1) % RM_MAXQUEUED;
		nmp->b_datap->db_type = mp->b_datap->db_type;
		nmp->b_cont = mp->b_cont;
		mp->b_cont = 0;
		nmp->b_wptr = mp->b_wptr;
		bp = nmp;
	} else
		bp = mp;

	cbqunlock(WR(q));
	putnext(WR(q), bp);
	cbqlock(WR(q));
}

/*
 * void
 * cbq_pktsched(void *arg, int nbytes)
 *	This function is called when a device driver has put a packet
 *	onto the wire.
 */
static void
cbq_pktsched(void *arg, int nbytes)
{
	cbqfree(arg, 0);
}

static void
cbq_esballoc_free(void *arg)
{
	cbqfree(arg, 0);
}

/*
 * void
 * cbqfree(cbq_module_t *, int)
 *
 *	Free the message header and start the next transfer.  This is
 *	called upon completion of the data transfer of the packet.
 *
 *	Returns:	NONE
 */

static void
cbqfree(void *arg, int flag)
{
	cbq_module_t	*cbqp = (cbq_module_t *)arg;
	mblk_t		*mp;

	cbqlock(WR(cbqp->cbq_q));
	if (cbqp->cbq_enabled) {
		/*
 	 	 * If we are using esballoc, then free the message.
 	 	 */
		if (cbqp->use_esballoc && cbqp->cbq_freeit[cbqp->cbqfo]) {
			freemsg(cbqp->cbq_freeit[cbqp->cbqfo]);
			cbqp->cbq_freeit[cbqp->cbqfo] = NULL;
			cbqp->cbqfo = (cbqp->cbqfo + 1) % RM_MAXQUEUED;
		}

		/* Update the class and schedule the next packet. */
		rmc_update_class_util(&cbqp->ifnp);
	
		/* Start the next packet, if there is one.  */
		cbq_deque(cbqp->cbq_q);
	}
	cbqunlock(WR(cbqp->cbq_q));
}

/*
 * int
 * cbq_deque(queue_t *) - Started sending of data.
 *
 *	Returns:	1, if a can be queued downstream.
 *			0, if a packet can't be queued.
 */

static int
cbq_deque(queue_t *q)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE(q->q_ptr);
	mblk_t		*mp;

	if (cbqp->cbq_busy)
		return (1);

	/*
	 * If we are inactive, then send out data if there is
	 * some to send.
	 */
	while (canputnext(WR(q))) {
		if ((mp = rmc_dequeue_next(&cbqp->ifnp)) != NULL) {
			mp->b_next = 0;
			cbqp->cbq_busy++;
			cbqstart(WR(q), mp, cbqp);
			cbqp->cbq_busy = 0;
		} else
			return (1);
	}

	return (0);
}

/*
 * void
 * cbqwatchdog(queue_t *q) - This function handles overlimit rm_class code
 *	callout.
 *
 *	Assumptions:	cbqlock has been called before calling this routine.
 *	Returns:	NONE
 */

void
cbqwatchdog(queue_t *q)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE(q->q_ptr);

	/*
	 * If the CBQ is enabled on this interface, make sure
	 * the it is running if there are packets to send.
	 */
	if (cbqp->cbq_enabled)
		cbq_deque(WR(q));
}

/*
 * void
 * cbqlock(queue_t *) - grab lock for this cbq_state_t struture.
 *
 *	Retunrns:	NONE
 */

void
cbqlock(queue_t *q)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE(q->q_ptr);
	mutex_enter(&cbqp->cbq_lock);
}

/*
 * void
 * cbqunlock(queue_t *) - let lock go
 *
 *	Returns:	NONE
 */

void
cbqunlock(queue_t *q)
{
	cbq_module_t	*cbqp = GET_CBQ_MODULE(q->q_ptr);
	mutex_exit(&cbqp->cbq_lock);
}
