/*
 * Copyright (c) Sun Microsystems, Inc. 1993-1998. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SMCC Technology
 *      Development Group at Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#pragma ident	"@(#)cbqddi.c	1.13	98/05/12 SMI"

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/kmem.h>

#include <netinet/rm_class.h>
#include <netinet/cbq.h>

#define	D_SD_COMMENT	"CBQ Streams Device v1.13"
#define	D_SD_INFO	cbqinfo
#define	D_SD_NAME	"cbq"
#define	D_SD_CTL_NAME	"cbqctl"
#define D_SD_DDI_TYPE	0
#define	D_SD_OPS_NAME	cbq_d_ops
#define D_SD_FLAGS	D_MP|D_MTPERMOD|D_MTPUTSHARED

/*
 * D_SD_FLAGS in the other tcp/ip modules have to these D_SD_FLAGS since they
 * are effectively clones of the ip driver with their module autopushed.
 */
#define	D_SM_COMMENT	"CBQ Streams Module v1.13"
#define	D_SM_INFO	cbqinfo
#define	D_SM_NAME	"cbq"
#define	D_SM_OPS_NAME	cbq_m_ops
#define D_SM_FLAGS	D_MP

/*
 * External Functions and variables.
 */
extern int		nulldev();
extern int		nodev();

extern int		cbq_ddi_init(dev_info_t *devi);
extern int		cbq_ddi_destroy(void);

#ifdef D_SD_INFO
extern struct mod_ops	mod_driverops;
extern struct streamtab	D_SD_INFO;

static int	_mi_driver_info(dev_info_t *dip, ddi_info_cmd_t infocmd,
			void *arg, void **result);
static int	_mi_driver_attach(dev_info_t * devi,  ddi_attach_cmd_t cmd);
static int	_mi_driver_detach(dev_info_t * devi,  ddi_detach_cmd_t cmd);
static int	_mi_driver_identify(dev_info_t * devi);

/* inter-module dependencies */
char _depends_on[] = "drv/ip";

static struct cb_ops _mi_driver_ops = {
	nulldev,		/* cb_open */
	nulldev,		/* cb_close */
	nodev,			/* cb_strategy */
	nodev,			/* cb_print */
	nodev,			/* cb_dump */
	nodev,			/* cb_read */
	nodev,			/* cb_write */
	nodev,			/* cb_ioctl */
	nodev,			/* cb_devmap */
	nodev,			/* cb_mmap */
	nodev,			/* cb_segmap */
	nochpoll,		/* cb_chpoll */
	ddi_prop_op,		/* cb_prop_op */
	&D_SD_INFO,		/* cb_stream */
	(int)D_SD_FLAGS		/* cb_flag */
};

struct dev_ops D_SD_OPS_NAME = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	_mi_driver_info,	/* devo_getinfo */
	_mi_driver_identify,	/* devo_identify */
	nulldev,		/* devo_probe */
	_mi_driver_attach,	/* devo_attach */
	_mi_driver_detach,	/* devo_detach */
	nodev,			/* devo_reset */
	&_mi_driver_ops,	/* devo_cb_ops */
	(struct bus_ops *)NULL	/* devo_bus_ops */
};

static struct modldrv modldrv = {
	&mod_driverops,	 D_SD_COMMENT, &D_SD_OPS_NAME
};
#endif

#ifdef D_SM_INFO
extern struct streamtab	D_SM_INFO;
extern struct mod_ops	mod_strmodops;

static int	_mi_module_info(dev_info_t *dip, ddi_info_cmd_t infocmd,
			void *arg, void **result);
static int	_mi_module_attach(dev_info_t *devi, ddi_attach_cmd_t cmd);
static int	_mi_module_identify(dev_info_t *devi);

static struct cb_ops _mi_module_ops = {
	nulldev,		/* cb_open */
	nulldev,		/* cb_close */
	nodev,			/* cb_strategy */
	nodev,			/* cb_print */
	nodev,			/* cb_dump */
	nodev,			/* cb_read */
	nodev,			/* cb_write */
	nodev,			/* cb_ioctl */
	nodev,			/* cb_devmap */
	nodev,			/* cb_mmap */
	nodev,			/* cb_segmap */
	nochpoll,		/* cb_chpoll */
	ddi_prop_op,		/* cb_prop_op */
	&D_SM_INFO,		/* cb_stream */
	(int)D_SM_FLAGS		/* cb_flag */
};

struct dev_ops D_SM_OPS_NAME = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	_mi_module_info,	/* devo_getinfo */
	_mi_module_identify,	/* devo_identify */
	nulldev,		/* devo_probe */
	_mi_module_attach,	/* devo_attach */
	nodev,			/* devo_detach */
	nodev,			/* devo_reset */
	&_mi_module_ops,	/* devo_cb_ops */
	(struct bus_ops *)NULL	/* devo_bus_ops */
};


static struct fmodsw _mi_module_fmodsw = {
	D_SM_NAME, &D_SM_INFO, D_SM_FLAGS
};

static struct modlstrmod modlstrmod = {
	&mod_strmodops, D_SM_COMMENT, &_mi_module_fmodsw
};
#endif

static struct modlinkage modlinkage = {
	MODREV_1, (void *)&modlstrmod, (void *)&modldrv, NULL
};

int
_init(void)
{
	return (mod_install(&modlinkage));
}

int
_fini(void)
{
	return (mod_remove(&modlinkage));
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

#ifdef D_SD_INFO
static	dev_info_t	* _mi_driver_dev_info;

/* ARGSUSED */
static int
_mi_driver_attach(dev_info_t *devi, ddi_attach_cmd_t cmd)
{
	_mi_driver_dev_info = devi;

	if (ddi_create_minor_node(devi, D_SD_NAME, S_IFCHR,
			ddi_get_instance(devi), NULL,
			D_SD_DDI_TYPE) == DDI_FAILURE) {
		cbq0dbg((CE_CONT,"_mi_driver_attach: fail\n"));
		ddi_remove_minor_node(devi, NULL);
		return (DDI_FAILURE);
	}

	if (ddi_create_minor_node(devi, D_SD_CTL_NAME, S_IFCHR,
			ddi_get_instance(devi) | CBQ_MINOR_CTL,
			NULL, D_SD_DDI_TYPE) == DDI_FAILURE) {
		cbq0dbg((CE_CONT,"_mi_driver_attach: fail\n"));
		ddi_remove_minor_node(devi, NULL);
		return (DDI_FAILURE);
	}

	if (!cbq_ddi_init(devi)) {
		cbq0dbg((CE_CONT,"_mi_driver_attach: fail\n"));
		ddi_remove_minor_node(devi, NULL);
		return (DDI_FAILURE);
	}

	return (DDI_SUCCESS);
}

static  int
_mi_driver_detach(dev_info_t *devi,  ddi_detach_cmd_t cmd)
{
	if (!cbq_ddi_destroy()) {
		cmn_err(CE_CONT,"cbq_ddi_destroy FAILED\n");
		return (DDI_FAILURE);
	}	

	return (DDI_SUCCESS);
}

/* ARGSUSED */
static int
_mi_driver_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
	void **result)
{
	register dev_t dev = (dev_t) arg;
	register int error;

	cbq0dbg((CE_CONT, "_mi_driver_info\n"));

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if (_mi_driver_dev_info == NULL) {
			error = DDI_FAILURE;
		} else {
			*result = (void *) _mi_driver_dev_info;
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)0;
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}

	cbq0dbg((CE_CONT,"_mi_driver_info\n"));
	return (error);
}


static int
_mi_driver_identify(dev_info_t *devi)
{
	cbq0dbg((CE_CONT, "_mi_driver_identify\n"));

	if (strcmp((char *)ddi_get_name(devi), D_SD_NAME) == 0) {
		cbq0dbg((CE_CONT, "_mi_driver_identify: idenitfied\n"));
		return (DDI_IDENTIFIED);
	}

	cbq0dbg((CE_CONT, "_mi_driver_identify: not idenitfied\n"));
	return (DDI_NOT_IDENTIFIED);
}
#endif

#ifdef D_SM_INFO
static	dev_info_t	* _mi_module_dev_info;

/* ARGSUSED */
static int
_mi_module_attach(dev_info_t *devi, ddi_attach_cmd_t cmd)
{
	cbq0dbg((CE_CONT, "_mi_module_attach\n"));

	_mi_module_dev_info = devi;

	cbq0dbg((CE_CONT,"_mi_module_attach\n"));
	return (DDI_SUCCESS);
}

/* ARGSUSED */
static int
_mi_module_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
	void **result)
{
	register dev_t dev = (dev_t) arg;
	register int error;

	cbq0dbg((CE_CONT, "_mi_module_info\n"));

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if (_mi_module_dev_info == NULL) {
			error = DDI_FAILURE;
		} else {
			*result = (void *) _mi_module_dev_info;
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)0;
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}

	cbq0dbg((CE_CONT,"_mi_module_info\n"));
	return (error);
}

static int
_mi_module_identify (devi)
	dev_info_t * devi;
{
	cbq0dbg((CE_CONT, "_mi_module_identify\n"));

	if (strcmp((char *)ddi_get_name(devi), D_SM_NAME) == 0) {
		cbq0dbg((CE_CONT, "_mi_module_identify: identified\n"));
		return (DDI_IDENTIFIED);
	}

	cbq0dbg((CE_CONT, "_mi_module_identify: not identified\n"));
	return (DDI_NOT_IDENTIFIED);
}
#endif
