/*
 * Copyright (c) Sun Microsystems, Inc. 1993-1998. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SMCC Technology
 *      Development Group at Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#pragma ident "@(#)pktsched.c  1.7     98/05/04 SMI"

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/strsun.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/dlpi.h>
#include <sys/stropts.h>
#include <sys/strlog.h>
#include <sys/tihdr.h>
#include <sys/time.h>
#include <sys/tiuser.h>
#include <sys/kstat.h>
#include <sys/kmem.h>

#include <sys/socket.h>
#include <net/route.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <inet/common.h>
#include <inet/mi.h>
#include <inet/ip.h>
#include <inet/ip_multi.h>
#include <inet/ip_ire.h>
#include <netinet/ip_mroute.h>
#include <sys/ethernet.h>

#include <netinet/pktsched.h>

#define PKTSCHED_TIMEOUT	100000

int
pktsched_init(queue_t *q, pkt_sched_cb_t *cb, int *flag,
	kmutex_t *mtx, kcondvar_t *cv)
{
	mblk_t		*mp;
	struct iocblk	*iocp;

	if ((mp = allocb(sizeof (struct iocblk), BPRI_HI)) == NULL)
		return (0);
	if ((mp->b_cont = allocb(sizeof (pkt_sched_cb_t), BPRI_HI)) == NULL)
		return (0);

	/*
	 * Set up the IOCTL to the driver.
	 */
	mp->b_datap->db_type = M_IOCTL;
	bzero((char *)mp->b_wptr, sizeof (*iocp));
	iocp = (struct iocblk *)mp->b_wptr;
	iocp->ioc_cmd = DL_IOC_PKTSCHED_INIT;
	iocp->ioc_count = msgdsize(mp->b_cont);
	mp->b_wptr += sizeof (*iocp);

	bcopy((void *)cb, (void *)mp->b_cont->b_wptr, sizeof (pkt_sched_cb_t));
	mp->b_cont->b_wptr += sizeof (pkt_sched_cb_t);

	/*
	 * Send the IOCTL.
	 */
	putq(WR(q), mp);
	qenable(q);

	/*
	 * Wait for response.
	 */
	while (*flag == 0)
		cv_wait(cv, mtx);

	if (*flag == -1)
		return (0);
	return (1);
}

int
pktsched_teardown(queue_t *q, int *flag, kmutex_t *mtx, kcondvar_t *cv)
{
	mblk_t		*mp;
	struct iocblk	*iocp;
	clock_t		lb;

	if ((mp = allocb(sizeof (struct iocblk), BPRI_HI)) == NULL)
		return (0);

	/*
	 * Build the ioctl.
	 */
	mp->b_datap->db_type = M_IOCTL;
	bzero((char *)mp->b_wptr, sizeof (*iocp));
	iocp = (struct iocblk *)mp->b_wptr;
	iocp->ioc_cmd = DL_IOC_PKTSCHED_TEARDOWN;
	iocp->ioc_count = 0;
	mp->b_wptr += sizeof (*iocp);

	/*
	 * Send the IOCTL.
	 */
	putq(WR(q), mp);
	qenable(q);

	/*
	 * Wait for response.
	 */
	drv_getparm(LBOLT, (unsigned long *)&lb);
	while (*flag == 0) {
		if (cv_timedwait(cv, mtx,
				lb + drv_usectohz(PKTSCHED_TIMEOUT)) < 0) {
			return (1);
		}
	}

	if (*flag == -1)
		return (0);
	return (1);
}
