/*
 * Copyright (c) Sun Microsystems, Inc. 1993-1998 All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SMCC Technology
 *      Development Group at Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" 
 * without express or implied warranty of any kind. 
 *  
 * These notices must be retained in any copies of any part of this software.
 */

#ifndef _NETINET_CBQ_H
#define	_NETINET_CBQ_H

#pragma ident "@(#)cbq.h  1.18     98/05/13 SMI"

#include <sys/ioctl.h>
#include <sys/ioccom.h>

#include <netinet/cbqio.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _KERNEL

/*
 * Define macros only good for kernel drivers and modules.
 */
#define	CBQ_DEVICE	0x1
#define	CBQ_MODULE	0x2

#define CBQ_MINOR_DEVICE	0x00
#define CBQ_MINOR_CTL		0x80
#define CBQ_ISCTL(minor)	(((minor) & CBQ_MINOR_CTL) == CBQ_MINOR_CTL)

/*
 * CBQ enable/disable macros
 */
#define	ENABLE		0x01
#define	DISABLE		0x00

/*
 * CBQ macros for fastpath processing
 */
#define	IS_FAST_PATH		0x01
#define	IS_NOT_FAST_PATH	0x00

/*
 * Stream definitions of hi and low water marks.
 */
#define	CBQ_HI_WATER		(256 * 1024)
#define	CBQ_LO_WATER		(1 * 1024)

/*
 * Other various marcos.
 */
#define	CBQ_WATCHDOG    	(HZ / 20)
#define	CBQ_TIMEOUT		1
#define	CBQ_LS_TIMEOUT		drv_usectohz(20000)
#define	CBQ_VIRTUAL_IF_THRESH	1000
#define	CBQ_ACTIVE		1
#define	CBQ_INACTIVE		0

#define	CBQ_MAX_CLASSES		256
#define	CBQ_MAX_FILTERS		256

/*
 * Define State structures.
 */
typedef struct _tbfilter_ {
	int		enable;	/* Is enabled */
	int		rate;	/* Bytes/millisecond */
	int		depth;	/* Bytes */
	int		bucket;	/* Bytes */
	hrtime_t	last;	/* Last timeval in microseconds. */
	u_long		dropped;/* Number of packets dropped */
} tbfilter_t;

typedef struct _k_filter_cb_ {
	struct rm_class		*clp;
	cbq_filt_t		filter;

	u_long			handle;

	struct _k_filter_cb_	*next;
	struct _k_filter_cb_	*prev;
	int			matches;

	tbfilter_t		tbfilter;
} k_filter_cb_t;

typedef struct cbq_module {
	kmutex_t		cbq_lock;
	kcondvar_t		cv;
#ifdef DEBUG
	int			cbqstarts;
	int			cbqfrees;
#endif
	queue_t			*cbq_q;

	int			mac_addr_sz;
	int			is_pktsched_inited;
	int			is_pktsched_destroyed;
	int			use_esballoc;
	int			cbq_timeout_id;
	int			cbq_tl_timeout_id;
	int			cbq_busy;
	int			cbq_enabled;
	int			qlen;
	int			q_full;
	u_long			threshold;

	int			cbqfi;
	int			cbqfo;
	mblk_t			*cbq_freeit[RM_MAXQUEUED];
	frtn_t			cbq_retn;

	struct rm_class		**cbq_class_tbl;

#ifndef MASK_AND_MATCH
#define	CBQ_FILTER_TBL_SZ	256
#define	CBQ_FILTER_MASK		(CBQ_FILTER_TBL_SZ - 1)
	k_filter_cb_t		*filters[CBQ_FILTER_TBL_SZ];
#else
	k_filter_cb_t		*filters;
	u_long			k_filt_handles;
#endif
	int			cbq_stats_timeout_id;
	int			cbq_stats_enabled;
	int			cbq_filter_debug;

	struct rm_ifdat		ifnp;
#ifdef CBQPERF
#define N_CBQ_PERF		64 * 1024
	struct cbq_perf		*perf;
	int			nperf;
#endif
} cbq_module_t;

typedef struct cbq_ill_s {
	char			cbq_ill_name[IFNAMSIZ];
	int			cbq_ill_name_length;
	int			cbq_ill_ppa;

	cbq_module_t		module;
} cbq_ill_t;

#ifndef MASK_AND_MATCH
#define	CBQ_GET_HASH_INDEX(addr) \
	(((addr) + (addr >> 8)) & (CBQ_FILTER_TBL_SZ - 1))
#define	CBQ_GET_HINDEX(handle) ((handle) >> 24)
#endif

typedef struct cbq_device {
	int			cbq_open;
	int			cbq_minor;
	queue_t			*cbq_q;
	dev_info_t		*dinfop;
} cbq_device_t;

typedef struct cbq_state {
	int			cbq_id;
	struct cbq_state	*cbq_next;
	union {
		cbq_ill_t	cbq_ms;
		cbq_device_t	cbq_ds;
	} cbq;
} cbq_state_t;

#define CBQ_NEXT(msp)		((cbq_state_t *)msp)->cbq_next
#define GET_CBQ_MODULE(msp)	&((cbq_state_t *)msp)->cbq.cbq_ms.module
#define GET_CBQ_DEVICE(msp)	&((cbq_state_t *)msp)->cbq.cbq_ds
#define CBQ_ILL_PPA(msp)	((cbq_state_t *)msp)->cbq.cbq_ms.cbq_ill_ppa
#define CBQ_ILL_NAME(msp)	((cbq_state_t *)msp)->cbq.cbq_ms.cbq_ill_name
#define CBQ_ILL_NAMELEN(msp) \
	((cbq_state_t *)msp)->cbq.cbq_ms.cbq_ill_name_length

int		cbqenque(queue_t *, mblk_t *, int);
void		cbqstart(queue_t *, mblk_t *, cbq_module_t *);
void		cbqwatchdog(queue_t *);
void		cbqlock(queue_t *);
void		cbqunlock(queue_t *);
#ifdef CBQPERF
int		cbq_initperf(queue_t *q);
int		cbq_destroyperf(queue_t *q);
void		cbq_perf(queue_t *q, int class_id,
			hrtime_t schedtime,
			hrtime_t departime);
#endif

int		cbq_class_destroy(cbq_module_t *, struct rm_class *);
struct rm_class *cbq_class_create(cbq_module_t *, struct cbq_add_class *,
			u_long, struct rm_class *, struct rm_class *);
int		cbq_class_queue(queue_t *, mblk_t *, int);

#ifdef CBQ_DEBUG
extern int cbq_debug;

#define cbq0dbg(a) 	if (cbq_debug > 1) cmn_err a;
#define cbq1dbg(a) 	if (cbq_debug > 2) cmn_err a;
#define cbq2dbg(a) 	if (cbq_debug > 3) cmn_err a;
#define cbq3dbg(a) 	if (cbq_debug > 4) cmn_err a;
#define cbq4dbg(a) 	if (cbq_debug > 5) cmn_err a;

#else CBQ_DEBUG

#define cbq0dbg(a)	/* */
#define cbq1dbg(a)	/* */
#define cbq2dbg(a)	/* */
#define cbq3dbg(a)	/* */
#define cbq4dbg(a)	/* */

#define	ENTRY(x)	/* */
#define	EXIT(x)		/* */

#endif /* CBQ_DEBUG */

#endif /* _KERNEL */

#ifdef __cplusplus
}
#endif

#endif /* !_NETINET_CBQ_H */
